/*
dttools is Copyright (C) 2002 Douglas Thain
This software is released under the GNU General Public License.
See the file COPYING for details.
*/


#include "safe_list.h"
#include "list.h"
#include "sleeptools.h"

#include <stdlib.h>
#include <pthread.h>

struct safe_list {
	struct list *list;
	pthread_mutex_t mutex;
	pthread_cond_t cond;
};

struct safe_list * safe_list_create()
{
	struct safe_list *l;

	l = malloc(sizeof(struct safe_list));
	if(!l) return 0;

	l->list = list_create();
	if(!l->list) {
		free(l);
		return 0;
	}

	pthread_mutex_init(&l->mutex,0);
	pthread_cond_init(&l->cond,0);

	return l;
}

void safe_list_delete( struct safe_list *l )
{
	list_delete(l->list);
	free(l);
}

 
void safe_list_push_head( struct safe_list *l, void *item )
{
	pthread_mutex_lock(&l->mutex);
	while(1) {
		if(list_push_head(l->list,item)) {
			pthread_cond_broadcast(&l->cond);
			break;
		} else {
			struct timespec ts;
			ts.tv_sec = time(0)+1;
			ts.tv_nsec = 0;
			pthread_cond_timedwait(&l->cond,&l->mutex,&ts);
		}
	}
	pthread_mutex_unlock(&l->mutex);
}

void safe_list_push_tail( struct safe_list *l, void *item )
{
	pthread_mutex_lock(&l->mutex);
	while(1) {
		if(list_push_tail(l->list,item)) {
			pthread_cond_broadcast(&l->cond);
			break;
		} else {
			struct timespec ts;
			ts.tv_sec = time(0)+1;
			ts.tv_nsec = 0;
			pthread_cond_timedwait(&l->cond,&l->mutex,&ts);
		}
	}
	pthread_mutex_unlock(&l->mutex);
}

void * safe_list_pop_head( struct safe_list *l )
{
	void *result;
	pthread_mutex_lock(&l->mutex);
	while(1) {
		result = list_pop_head(l->list);
		if(result) {
			break;
		} else {
			pthread_cond_wait(&l->cond,&l->mutex);
		}
	}
	pthread_cond_broadcast(&l->cond);
	pthread_mutex_unlock(&l->mutex);
	return result;
}

void * safe_list_pop_tail( struct safe_list *l )
{
	void *result;
	pthread_mutex_lock(&l->mutex);
	while(1) {
		result = list_pop_tail(l->list);
		if(result) {
			break;
		} else {
			pthread_cond_wait(&l->cond,&l->mutex);
		}
	}
	pthread_cond_broadcast(&l->cond);
	pthread_mutex_unlock(&l->mutex);
	return result;
}

void safe_list_flush( struct safe_list *l )
{
	pthread_mutex_lock(&l->mutex);
	while(list_size(l->list)>0) {
		pthread_cond_wait(&l->cond,&l->mutex);
	}
	pthread_mutex_unlock(&l->mutex);
}
