/*
dttools is Copyright (C) 2002 Douglas Thain
This software is released under the GNU General Public License.
See the file COPYING for details.
*/


#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "timer.h"
#include "hash_table.h"
#include "macros.h"
#include "debug.h"

struct event {
	timer_callback_t callback;
	void *arg;
	time_t last_call;
	int interval;
	int call_in_progress;
};

static struct hash_table *event_table = 0;
static time_t current_time=0;
static time_t next_call=0;

#ifdef sgi
static pthread_mutex_t mutex = { PTHREAD_MUTEX_INITIALIZER };
static pthread_cond_t cond = { PTHREAD_COND_INITIALIZER };
#else
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
#endif

void * timer_launch( void * arg )
{
	struct event *e = arg;

	e->callback(e->arg);
	e->call_in_progress = 0;

	pthread_mutex_lock(&mutex);
	pthread_cond_broadcast(&cond);	
	pthread_mutex_unlock(&mutex);

	return 0;
}

static int timer_dispatch( struct hash_table *h, char *key, void *value, void *arg )
{
	struct event *e = value;
	pthread_t tid;

	if( !e->call_in_progress ) {
		if( (e->last_call+e->interval) <= current_time) {
			debug("timer: dispatching %s",key);
			e->call_in_progress = 1;
			e->last_call = current_time;
			pthread_create(&tid,0,timer_launch,e);
			pthread_detach(tid);
		} else if( (e->last_call+e->interval) < next_call) {
			next_call = (e->last_call+e->interval);
		}
	}

	return 1;
}

static void * timer_thread( void *arg )
{
	struct timespec ts;

	pthread_mutex_lock(&mutex);
	while(1) {
		current_time = time(0);
		next_call = current_time+60;
		hash_table_iterate( event_table, timer_dispatch, 0 );
		ts.tv_sec = next_call;
		ts.tv_nsec = 0;
		pthread_cond_timedwait(&cond,&mutex,&ts);
	}
	pthread_mutex_unlock(&mutex);
	return 0;
}

static int timer_init()
{
	int save_errno;

	if(!event_table) {
		pthread_t tid;

		event_table = hash_table_create(127,hash_string);
		if(!event_table) goto failure;

		if(pthread_create(&tid,0,timer_thread,0)) goto failure;
		pthread_detach(tid);
	}

	return 1;

	failure:
	save_errno = errno;
	if(event_table) hash_table_delete(event_table);
	errno = save_errno;
	return 0;
}

int timer_add( const char *key, timer_callback_t callback, void *arg, int interval )
{
	struct event *old, *e=0;
	int save_errno;

	pthread_mutex_lock(&mutex);

	if(!timer_init()) goto failure;

	e = malloc(sizeof(*e));
	if(!e) goto failure;

	e->callback = callback;
	e->arg = arg;
	e->last_call = 0;
	e->interval = interval;
	e->call_in_progress = 0;

	if(!hash_table_insert(event_table,key,e,(void**)&old)) goto failure;

	if(old) free(old);

	debug("timer: added %s interval %d",key,interval);

	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&mutex);	
	return 1;

	failure:
	save_errno = errno;
	if(e) free(e);
	pthread_mutex_unlock(&mutex);
	errno = save_errno;
	return 0;
}

int timer_remove( const char *key )
{
	struct event *e;
	int result;

	pthread_mutex_lock(&mutex);

	if(timer_init()) {
		e = hash_table_remove(event_table,key);
		if(e) {
			free(e);
			result = 1;
		} else {
			result = 0;
		}
	} else {
		result = 0;
	}
	pthread_mutex_unlock(&mutex);

	return result;
}

